#!/usr/bin/perl -w
#
# jstockd -- the JStock Daemon
#
# This daemon is a UDP server that
# serves the information retrieved
# and formatted by priced to jstock
# clients.
#

require 5.004;
use POSIX;
use IO::Socket;
use Symbol;

sub REAPER; # bury dead children :-)
sub HUNTER; # sigint handler--kill our children
sub make_new_child; # make a child process

# GLOBAL VARIABLES #

$SERVER_PORT = 3130;
$PREFORK = 5;
%children = ();
$children = 0;
$MAX_CLIENTS_PER_CHILD = 5;

# /GLOBAL VARIABLES #

## DAEMONIZE ##
# When we are run, the first
# thing to do is detach from the
# terminal and become a daemon.
# Then we run in the background.
 
# change working directory
chdir("/var/jstock")
    or die "Could not chdir to /var/jstock: $!\n";

# open the logfile
open (LOG, ">jstockd.log")
    or die "Couldn't open log file: $!\n";

# redirect stdout and stderr
#open (STDOUT, ">&LOG") or die "Could not redirect STDOUT: $!\n";
#open (STDERR, ">&LOG") or die "Could not redirect STDERR: $!\n";

select (LOG);
$|=1;
select (STDOUT);

$pid = fork;
if ($pid)
{
    open (PID, ">/var/run/jstockd.pid");
    print PID $pid, "\n";
    close PID;
    exit;
}
die "Couldn't fork: $!" unless defined($pid);

die "/var/jstock does not exist: $!\n" unless (-d "/var/jstock");

POSIX::setsid()
    or die "Can't start a new session: $!\n";

# get the server socket
$server = IO::Socket::INET->new(LocalPort => $SERVER_PORT,
				Type => SOCK_STREAM,
				Proto => 'tcp',
				Reuse => 1,
				Listen => 10)
    or die "Couldn't start TCP server on port $SERVER_PORT: $!\n";

$server->autoflush(1);

# premake all child processes
for (1 .. $PREFORK)
{
    make_new_child();
}

$SIG{CHLD} = \&REAPER;
$SIG{INT} = $SIG{KILL} = $SIG{TERM} = \&HUNTER;

while (1)
{
    sleep; # wait for a signal such as child death or termination
    for ($i = $children; $i < $PREFORK; $i++)
    {
	make_new_child(); # keep lots of kiddies alive
    }
}

## BEGIN SUB MAKE_NEW_CHILD ##

sub make_new_child
{
    my $pid;
    my $sigset;

    # block SIGINT for the fork
    $sigset = POSIX::SigSet->new(SIGINT);
    sigprocmask(SIG_BLOCK, $sigset)
	or die "Can't block SIGINT for fork: $!\n";

    die "Couldn't fork: $!\n" unless defined ($pid = fork);

    if ($pid)
    {
	# Parent loop--record child and return

	sigprocmask(SIG_UNBLOCK, $sigset)
	    or die "Can't unblock SIGINT after fork: $!\n";

	$children{$pid} = 1;
	$children++;
	return;
    }
    else
    {
	# Child loop--the real server magic happens here
	
	# make signals copacetic
	$SIG{INT} = 'DEFAULT';
	sigprocmask(SIG_UNBLOCK, $sigset)
	    or die "Can't unblock SIGINT after fork in child: $!\n";
	
	# handle connections until we reach our preassigned max
	for ($i = 0; $i < $MAX_CLIENTS_PER_CHILD; $i++)
	{
	    $client = $server->accept() or last;
	    # do our magic

	    # open info file
	    open (INFO, "jstock.info") 
		or die "Couldn't open info file: $!\n";
	    
	    # read lines from info file
	    $line = <INFO>;
	    $line .= <INFO>;
	    $line .= <INFO>;
	    
	    # close info file
	    close INFO;

	    # turn newlines into a record seperator, \0
	    $line =~ s/\n/\0/gs;

	    print LOG $line;
	    
	    # send the data
	    $client->send($line);
	}
	
	exit;

    }

}

## ENDOF SUB MAKE_NEW_CHILD ##

## BEGIN SUB REAPER ##

sub REAPER
{
    $SIG{CHLD} = \&REAPER;
    my $pid = wait;
    $children--;
    delete $children{$pid};
}

## ENDOF SUB REAPER ##

## BEGIN SUB HUNTER ##

sub HUNTER
{
    local($SIG{CHLD}) = 'IGNORE';
    kill ('INT', keys %children);
    wait;
    exit;
}

## ENDOF SUB HUNTER ##
